/*
 * Copyright 2023 Alexey Lyashko
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the “Software”), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is furnished
 * to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED “AS IS”, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 * INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 * PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 * SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */


#include "include/trimp.h"
#include <stdio.h>
#include <math.h>

#define NET_RATE 0.01
#define NET_MOMENTUM 0.0002
#define ERROR_LIMIT 0.000001

#define USE_DIFFERENT_ACTIVATION_PER_NEURON 1
#define RECURRENT 1

double get_mean_error(const double *errors, int size)
{
    double me = 0;
    for (int i = 0; i < size; i++)
    {
        me += *(errors + i) / 2.0;
    }
    return me / (double) size;
}

int main()
{
    int neuronsPerLayer[] = {2, 3, 1};

    double inputData[][2] = {
            {1.0, 1.0},
            {1.0, 0.0},
            {0.0, 1.0},
            {0.0, 0.0}
    };

    double expected[][1] = {
            {0.0},
            {1.0},
            {1.0},
            {0.0}
    };

    double outputData[4];

    double error[4];

#ifdef USE_DIFFERENT_ACTIVATION_PER_NEURON
    enum activation_type layer1[] = {Exponential, Exponential};
    enum activation_type layer2[] = {Exponential, HTangent, Gauss};
    enum activation_type layer3[] = {Gauss};
    enum activation_type *activationType[] = {layer1, layer2, layer3};

    pnet_t net = Net_new_ex(sizeof neuronsPerLayer / sizeof(int), neuronsPerLayer, activationType);
#else
    enum activation_type activationType[]={Exponential, HTangent, Gauss};
    pnet_t net = Net_new(sizeof neuronsPerLayer / sizeof(int), neuronsPerLayer, activationType);
#endif

    NetSetRate(net, NET_RATE);
    NetSetMomentum(net, NET_MOMENTUM);

#ifdef RECURRENT
    NetSetType(net, Elman);
#endif


    for (int j = 0; j < 100000; j++)
    {
        double meanError;
        for (int i = 0; i < 4; i++)
        {
            error[i] = NetTrain(net, inputData[i], &outputData[i], expected[i]);
        }
        meanError = get_mean_error(error, 4);
        if (j % 10000 == 0 || meanError < ERROR_LIMIT)
            fprintf(stderr, "\rEpoch: %10d;   Error: %1.25f\t\tValues: %1.25f\t%1.25f\t%1.25f\t%1.25f", j, meanError,
                    outputData[0],
                    outputData[1], outputData[2], outputData[3]);
        if (meanError < ERROR_LIMIT)
            break;

        if (isnan(meanError))
        {
            fprintf(stderr, "\n\n*** The output value is not a number. ***\n");
            break;
        }
    }
    printf("\n");
    Net_free(net);
    return 0;
}